@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2017-2018 Joris Vink <joris@coders.se>
+ * Copyright (c) 2017-2019 Joris Vink <joris@coders.se>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -21,6 +21,7 @@ struct python_coro {
 	u_int64_t			id;
 	int				state;
 	PyObject			*obj;
+	PyObject			*result;
 	struct pysocket_op		*sockop;
 	struct pygather_op		*gatherop;
 	struct http_request		*request;
@@ -40,18 +41,23 @@ static PyObject		*python_kore_timer(PyObject *, PyObject *);
 static PyObject		*python_kore_fatal(PyObject *, PyObject *);
 static PyObject		*python_kore_queue(PyObject *, PyObject *);
 static PyObject		*python_kore_tracer(PyObject *, PyObject *);
-static PyObject		*python_kore_gather(PyObject *, PyObject *);
 static PyObject		*python_kore_fatalx(PyObject *, PyObject *);
 static PyObject		*python_kore_suspend(PyObject *, PyObject *);
 static PyObject		*python_kore_shutdown(PyObject *, PyObject *);
 static PyObject		*python_kore_bind_unix(PyObject *, PyObject *);
 static PyObject		*python_kore_task_create(PyObject *, PyObject *);
 static PyObject		*python_kore_socket_wrap(PyObject *, PyObject *);
+static PyObject		*python_kore_gather(PyObject *, PyObject *, PyObject *);
 
 #if defined(KORE_USE_PGSQL)
 static PyObject		*python_kore_pgsql_register(PyObject *, PyObject *);
 #endif
 
+#if defined(KORE_USE_CURL)
+static PyObject		*python_kore_httpclient(PyObject *,
+			    PyObject *, PyObject *);
+#endif
+
 static PyObject		*python_websocket_broadcast(PyObject *, PyObject *);
 
 #define METHOD(n, c, a)		{ n, (PyCFunction)c, a, NULL }
@@ -68,7 +74,7 @@ static struct PyMethodDef pykore_methods[] = {
 	METHOD("timer", python_kore_timer, METH_VARARGS),
 	METHOD("queue", python_kore_queue, METH_VARARGS),
 	METHOD("tracer", python_kore_tracer, METH_VARARGS),
-	METHOD("gather", python_kore_gather, METH_VARARGS),
+	METHOD("gather", python_kore_gather, METH_VARARGS | METH_KEYWORDS),
 	METHOD("fatal", python_kore_fatal, METH_VARARGS),
 	METHOD("fatalx", python_kore_fatalx, METH_VARARGS),
 	METHOD("suspend", python_kore_suspend, METH_VARARGS),
@@ -79,6 +85,10 @@ static struct PyMethodDef pykore_methods[] = {
 	METHOD("websocket_broadcast", python_websocket_broadcast, METH_VARARGS),
 #if defined(KORE_USE_PGSQL)
 	METHOD("register_database", python_kore_pgsql_register, METH_VARARGS),
+#endif
+#if defined(KORE_USE_CURL)
+	METHOD("httpclient", python_kore_httpclient,
+	    METH_VARARGS | METH_KEYWORDS),
 #endif
 	{ NULL, NULL, 0, NULL }
 };
@@ -147,13 +157,28 @@ static PyTypeObject pytimer_type = {
 	.tp_flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,
 };
 
+/* XXX */
+struct pysocket;
+struct pysocket_op;
+
+struct pysocket_event {
+	struct kore_event	evt;
+	struct pysocket		*s;
+};
+
 struct pysocket {
 	PyObject_HEAD
 	int			fd;
 	int			family;
 	int			protocol;
+	int			scheduled;
 	PyObject		*socket;
 	socklen_t		addr_len;
+
+	struct pysocket_event	event;
+	struct pysocket_op	*recvop;
+	struct pysocket_op	*sendop;
+
 	union {
 		struct sockaddr_in	ipv4;
 		struct sockaddr_un	sun;
@@ -198,23 +223,22 @@ static PyTypeObject pysocket_type = {
 #define PYSOCKET_TYPE_RECVFROM	5
 #define PYSOCKET_TYPE_SENDTO	6
 
-struct pysocket_data {
-	struct kore_event	evt;
-	int			fd;
-	int			eof;
-	int			type;
-	void			*self;
-	struct python_coro	*coro;
-	int			state;
-	size_t			length;
-	struct kore_buf		buffer;
-	struct sockaddr_in	sendaddr;
-	struct pysocket		*socket;
-};
-
 struct pysocket_op {
 	PyObject_HEAD
-	struct pysocket_data	data;
+	int				eof;
+	int				type;
+	void				*self;
+	struct python_coro		*coro;
+	int				state;
+	size_t				length;
+	struct kore_buf			buffer;
+	struct pysocket			*socket;
+	struct kore_timer		*timer;
+
+	union {
+		struct sockaddr_in	ipv4;
+		struct sockaddr_un	sun;
+	} sendaddr;
 };
 
 static void	pysocket_op_dealloc(struct pysocket_op *);
@@ -445,6 +469,8 @@ struct pygather_result {
 struct pygather_op {
 	PyObject_HEAD
 	int				count;
+	int				running;
+	int				concurrency;
 	struct python_coro		*coro;
 	TAILQ_HEAD(, pygather_result)	results;
 	TAILQ_HEAD(, pygather_coro)	coroutines;
@@ -521,6 +547,13 @@ struct pyhttp_request {
 	PyObject		*data;
 };
 
+struct pyhttp_iterobj {
+	int			remove;
+	PyObject		*iterator;
+	struct connection	*connection;
+	struct kore_buf		buf;
+};
+
 struct pyhttp_file {
 	PyObject_HEAD
 	struct http_file	*file;
@@ -530,7 +563,7 @@ static void	pyhttp_dealloc(struct pyhttp_request *);
 static void	pyhttp_file_dealloc(struct pyhttp_file *);
 
 #if defined(KORE_USE_PGSQL)
-static PyObject	*pyhttp_pgsql(struct pyhttp_request *, PyObject *);
+static PyObject	*pyhttp_pgsql(struct pyhttp_request *, PyObject *, PyObject *);
 #endif
 static PyObject *pyhttp_cookie(struct pyhttp_request *, PyObject *);
 static PyObject	*pyhttp_response(struct pyhttp_request *, PyObject *);
@@ -548,7 +581,7 @@ static PyObject *pyhttp_websocket_handshake(struct pyhttp_request *,
 
 static PyMethodDef pyhttp_request_methods[] = {
 #if defined(KORE_USE_PGSQL)
-	METHOD("pgsql", pyhttp_pgsql, METH_VARARGS),
+	METHOD("pgsql", pyhttp_pgsql, METH_VARARGS | METH_KEYWORDS),
 #endif
 	METHOD("cookie", pyhttp_cookie, METH_VARARGS),
 	METHOD("response", pyhttp_response, METH_VARARGS),
@@ -622,6 +655,88 @@ static PyTypeObject pyhttp_file_type = {
 	.tp_flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,
 };
 
+#if defined(KORE_USE_CURL)
+struct pyhttp_client {
+	PyObject_HEAD
+	char			*url;
+	char			*unix;
+	char			*tlskey;
+	char			*tlscert;
+	char			*cabundle;
+	int			tlsverify;
+};
+
+#define PYHTTP_CLIENT_OP_RUN	1
+#define PYHTTP_CLIENT_OP_RESULT	2
+
+struct pyhttp_client_op {
+	PyObject_HEAD
+	int			state;
+	int			headers;
+	struct kore_curl	curl;
+	struct python_coro	*coro;
+};
+
+static PyObject	*pyhttp_client_op_await(PyObject *);
+static PyObject	*pyhttp_client_op_iternext(struct pyhttp_client_op *);
+
+static void	pyhttp_client_dealloc(struct pyhttp_client *);
+static void	pyhttp_client_op_dealloc(struct pyhttp_client_op *);
+
+static PyObject *pyhttp_client_get(struct pyhttp_client *,
+		    PyObject *, PyObject *);
+static PyObject *pyhttp_client_put(struct pyhttp_client *,
+		    PyObject *, PyObject *);
+static PyObject *pyhttp_client_post(struct pyhttp_client *,
+		    PyObject *, PyObject *);
+static PyObject *pyhttp_client_head(struct pyhttp_client *,
+		    PyObject *, PyObject *);
+static PyObject *pyhttp_client_patch(struct pyhttp_client *,
+		    PyObject *, PyObject *);
+static PyObject *pyhttp_client_delete(struct pyhttp_client *,
+		    PyObject *, PyObject *);
+static PyObject *pyhttp_client_options(struct pyhttp_client *,
+		    PyObject *, PyObject *);
+
+static PyMethodDef pyhttp_client_methods[] = {
+	METHOD("get", pyhttp_client_get, METH_VARARGS | METH_KEYWORDS),
+	METHOD("put", pyhttp_client_put, METH_VARARGS | METH_KEYWORDS),
+	METHOD("post", pyhttp_client_post, METH_VARARGS | METH_KEYWORDS),
+	METHOD("head", pyhttp_client_head, METH_VARARGS | METH_KEYWORDS),
+	METHOD("patch", pyhttp_client_patch, METH_VARARGS | METH_KEYWORDS),
+	METHOD("delete", pyhttp_client_delete, METH_VARARGS | METH_KEYWORDS),
+	METHOD("options", pyhttp_client_options, METH_VARARGS | METH_KEYWORDS),
+	METHOD(NULL, NULL, -1)
+};
+
+static PyTypeObject pyhttp_client_type = {
+	PyVarObject_HEAD_INIT(NULL, 0)
+	.tp_name = "kore.httpclient",
+	.tp_doc = "An asynchronous HTTP client",
+	.tp_methods = pyhttp_client_methods,
+	.tp_basicsize = sizeof(struct pyhttp_client),
+	.tp_dealloc = (destructor)pyhttp_client_dealloc,
+	.tp_flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,
+};
+
+static PyAsyncMethods pyhttp_client_op_async = {
+	(unaryfunc)pyhttp_client_op_await,
+	NULL,
+	NULL
+};
+
+static PyTypeObject pyhttp_client_op_type = {
+	PyVarObject_HEAD_INIT(NULL, 0)
+	.tp_name = "kore.httpclientop",
+	.tp_doc = "Asynchronous HTTP client operation",
+	.tp_as_async = &pyhttp_client_op_async,
+	.tp_iternext = (iternextfunc)pyhttp_client_op_iternext,
+	.tp_basicsize = sizeof(struct pyhttp_client_op),
+	.tp_dealloc = (destructor)pyhttp_client_op_dealloc,
+	.tp_flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,
+};
+#endif
+
 #if defined(KORE_USE_PGSQL)
 
 #define PYKORE_PGSQL_PREINIT		1
@@ -632,15 +747,25 @@ static PyTypeObject pyhttp_file_type = {
 struct pykore_pgsql {
 	PyObject_HEAD
 	int			state;
-	char			*db;
-	char			*query;
-	struct http_request	*req;
-	PyObject		*result;
+	int			binary;
 	struct kore_pgsql	sql;
+
+	char			*db;
+	struct http_request	*req;
+	char			*query;
+	PyObject		*result;
+	struct {
+		int		count;
+		const char	**values;
+		int		*lengths;
+		int		*formats;
+		PyObject	**objs;
+	} param;
 };
 
 static void	pykore_pgsql_dealloc(struct pykore_pgsql *);
-int		pykore_pgsql_result(struct pykore_pgsql *);
+static int	pykore_pgsql_result(struct pykore_pgsql *);
+static int	pykore_pgsql_params(struct pykore_pgsql *, PyObject *);
 
 static PyObject	*pykore_pgsql_await(PyObject *);
 static PyObject	*pykore_pgsql_iternext(struct pykore_pgsql *);
